import yaml
import os

SETTINGS_FILES = [
    os.path.join(os.path.dirname(__file__), 'config.yaml'),
    os.path.expanduser('~/.esahub.conf')
]
CONFIG = {}


#
# Settings files in order of increasing precedence
#
def load(fname):
    fname = os.path.expanduser(fname)
    if os.path.isfile(fname):
        with open(fname, 'r') as fid:
            CONFIG.update(yaml.load(fid))
    CONFIG['GENERAL']['DATA_DIR'] = \
        os.path.expanduser(CONFIG['GENERAL']['DATA_DIR'])


for f in SETTINGS_FILES:
    load(f)



from esahub import config, utils, scihub
import os
import sys
import shutil

# -----------------------------------------------------------------------------
# TEST CONFIGURATION
# -----------------------------------------------------------------------------

#
# WARNING: Make sure the content of the following directories can be safely
# deleted!
# If they don't exist, they will be created.
#
TEST_DATA_DIR_ORIGINAL = 'esahub/tests/data/original/'
TEST_DATA_DIR_CORRUPT = 'esahub/tests/data/corrupt/'
TEST_DATA_DIR_TMP = 'esahub/tests/data/tmp/'


def _corrupt_binary(path):
    with open(path, 'wb') as f:
        f.write(b'78sadb')


def prepare():
    #
    # Create data directory structure
    #
    for d in (TEST_DATA_DIR_ORIGINAL, TEST_DATA_DIR_CORRUPT,
              TEST_DATA_DIR_TMP, ):
        if not os.path.exists(d):
            os.makedirs(d)
    #
    # Download test data
    #
    query = {'query': 'size: ???.* KB',
             'sort': ('size', 'asc'),
             'server': 'all'}
    fs = scihub.search(query, limit=2)
    _config = config.CONFIG['GENERAL']['DATA_DIR']
    config.CONFIG['GENERAL']['DATA_DIR'] = TEST_DATA_DIR_ORIGINAL
    scihub.download(fs)
    # move and corrupt one of the files
    corrupt_file = os.listdir(TEST_DATA_DIR_ORIGINAL)[-1]
    _move_from = os.path.join(TEST_DATA_DIR_ORIGINAL, corrupt_file)
    _move_to = os.path.join(TEST_DATA_DIR_CORRUPT, corrupt_file)
    shutil.move(_move_from, _move_to)
    _corrupt_binary(_move_to)
    config.CONFIG['GENERAL']['DATA_DIR'] = _config


def cleanup():
    clear_dirs = [TEST_DATA_DIR_ORIGINAL, TEST_DATA_DIR_CORRUPT]
    for d in clear_dirs:
        # Safety check
        if 'test' not in d:
            print('EXITING')
            sys.exit()

        # Delete all files in the directory
        for f in os.listdir(d):
            os.remove(os.path.join(d, f))


def set_test_config():
    # logging.disable(logging.CRITICAL)
    # Overwrite custom settings in ~/.esahub.conf
    config.load('esahub/config.yaml')
    config.CONFIG['GENERAL']['QUIET'] = True
    config.CONFIG['GENERAL']['DATA_DIR'] = TEST_DATA_DIR_TMP
    config.CONFIG['GENERAL']['TIMEOUT'] = 10.0
    config.CONFIG['GENERAL']['RECONNECT_TIME'] = 0.0
    config.CONFIG['GENERAL']['TRIALS'] = 3
    config.CONFIG['GENERAL']['WAIT_ON_503'] = False


def copy_test_data():
    for f in utils.ls(TEST_DATA_DIR_ORIGINAL):
        target = os.path.join(config.CONFIG['GENERAL']['DATA_DIR'],
                              os.path.split(f)[1])
        shutil.copy2(f, target)


def copy_corrupt_data():
    for f in utils.ls(TEST_DATA_DIR_CORRUPT):
        target = os.path.join(config.CONFIG['GENERAL']['DATA_DIR'],
                              os.path.split(f)[1])
        shutil.copy2(f, target)


def clear_test_data(tmp_only=False):
    clear_dirs = [config.CONFIG['GENERAL']['DATA_DIR']]
    for d in clear_dirs:
        # Safety check
        if 'test' not in d:
            print('EXITING')
            sys.exit()

        # Delete all files in the directory
        for f in os.listdir(d):
            os.remove(os.path.join(d, f))


def clear_all():
    clear_test_data()



# coding=utf-8
from __future__ import print_function
import os
import sys
import datetime as DT
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse, parserinfo
from dateutil.tz import tzutc
import re
import json
from distutils.spawn import find_executable
from collections import OrderedDict


PY2 = sys.version_info < (3, 0)
DATE_FMT = '%Y-%m-%dT%H:%M:%SZ'


# -----------------------------------------------------------------------------
# Python 2/3 compatibility.
# -----------------------------------------------------------------------------
def _which(exe):
    """Returns path of `exe` if found, else None."""
    return find_executable(exe)


_range = range if not PY2 else xrange


def eprint(*args, **kwargs):
    """Print to stderr"""
    print(*args, file=sys.stderr, **kwargs)


# -----------------------------------------------------------------------------
# Datetime parsing and formatting
# -----------------------------------------------------------------------------
def parse_datetime(value, dayfirst=False, force=None):
    """
    Parse a date time string.

    Parameters
    ----------
    force : str, optional
        - 'date' : Force return of a single datetime object.
        - 'range' : Force return of a range of datetime objects.

    Returns
    -------

    """
    def _parse_single_value(value):
        # Only extract the datetime parts that are actually given!
        defaults = [DT.datetime(1, 1, 1, 0, 0, 0),
                    DT.datetime(2, 2, 2, 2, 2, 2)]
        info = parserinfo(dayfirst=dayfirst)
        parsed = [parse(value, info, default=d) for d in defaults]
        start = parsed[0]

        if start == parsed[1]:
            # All attributes have been provided,
            # i.e. an exact date has been passed.
            return (start, start)

        # Determine which attributes have actually been specified.
        attrs = ['year', 'month', 'day', 'hour', 'minute', 'second']
        given_attrs = set(
            [attr for attr in attrs if
             getattr(parsed[1], attr) == getattr(parsed[0], attr)])

        # If some attributes are left out, return a time span.
        if given_attrs == set(['year']):
            end = start + relativedelta(years=1)
        elif given_attrs == set(['year', 'month']):
            end = start + relativedelta(months=1)
        elif given_attrs == set(['year', 'month', 'day']):
            end = start + relativedelta(days=1)
        elif given_attrs == set(['year', 'month', 'day', 'hour']):
            end = start + relativedelta(hours=1)
        else:
            return (start, start)

        return (start, end)

    splitters = ['to', '-', ',']
    try:
        stripped = value.strip()
        # Check for open ended date range:
        for s in splitters:
            if stripped.startswith(s):
                parsed = _parse_single_value(stripped.lstrip(s))
                return (None, parsed[1])
            elif stripped.endswith(s):
                parsed = _parse_single_value(stripped.rstrip(s))
                return (parsed[0], None)
        # Treat as single date
        parsed = _parse_single_value(value)
        return parsed

    except ValueError as e:
        #
        # Attempt to split into pair of values
        # if explicitly given a range
        #
        for splitter in splitters:
            if splitter in value:
                parts = value.split(splitter)
                if len(parts) > 2:
                    limit = len(parts) // 2
                    parts = (splitter.join(parts[:limit]),
                             splitter.join(parts[limit:]))
                parts = [_.strip() for _ in parts]
                start, end = [_parse_single_value(_) if bool(_) else None
                              for _ in parts]
                start = start[0]
                end = end[1]
                return start, end

    raise ValueError("Could not parse as datetime: %s" % value)


def to_date(string, fmt=None, output='str'):
    """Date to string with forced timezone information."""
    if string.startswith('UTC='):
        string = string[4:] + 'Z'
    if fmt is None:
        date_object = parse(string)
    else:
        date_object = DT.datetime.strptime(string, fmt)
    if date_object.tzinfo is None:
        date_object = date_object.replace(tzinfo=tzutc())
    if output == 'str':
        # return date_object.isoformat()
        return date_object.strftime(DATE_FMT)
    elif output == 'date':
        return date_object


def fix_product_name(product):
    return os.path.split(product)[1]


def fix_processing_level(string):
    return string.replace('Level-', 'L').strip()


def nth_char(string, n, convert=str, fmt='{}'):
    return convert(fmt.format(string[n]))


def substring(string, a, b):
    return string[a:b]


def strip(string):
    return string.strip()


def get_satellite(filename):
    f = os.path.split(filename)[1]
    return f[:f.find('_')]


def level_from_filename(filename):
    sat = get_satellite(filename)
    f = os.path.split(filename)[1]
    if sat == 'AE':
        return f[16:18]


def duration_from_filename(filename):
    sat = get_satellite(filename)
    f = os.path.split(filename)[1]
    if sat == 'AE':
        return str(int(int(f[38:47].lstrip('0'))/1000.))


# -----------------------------------------------------------------------------
# Human readable byte sizes
# -----------------------------------------------------------------------------
def b2h(num, suffix='B'):
    """Format file sizes as human readable.

    https://stackoverflow.com/a/1094933

    Parameters
    ----------
    num : int
        The number of bytes.
    suffix : str, optional
        (default: 'B')

    Returns
    -------
    str
        The human readable file size string.

    Examples
    --------
    >>> b2h(2048)
    '2.0kB'
    """
    try:
        for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z']:
            if abs(num) < 1024.0:
                return '{0:3.1f}{1}{2}'.format(num, unit, suffix)
            num /= 1024.0
        return '{0:.1f}{1}{2}'.format(num, 'Y', suffix)
    except:
        return '-'


def h2b(hstr, suffix='B'):
    """Format file sizes from human readable to bytes.

    Parameters
    ----------
    hstr : str
        The human readable size string.
    suffix : str, optional
        (default: 'B')

    Returns
    -------
    float
        The number of bytes if successfully parsed. None otherwise.

    Examples
    --------
    >>> h2b('3.5MB')
    3670016.0
    """
    m = re.search('^[0-9\.]*', hstr)
    num = float(m.group(0))
    unit = hstr.replace(m.group(0), '').lstrip().upper()
    # factor = 1.0
    for prefix in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if unit == prefix + suffix:
            return num
        num *= 1024.0
    return None


# -----------------------------------------------------------------------------
# List and dict operations
# -----------------------------------------------------------------------------
def flatten(l):
    """Flatten a list of lists.

    Parameters
    ----------
    l : list of lists
        A nested list

    Returns
    -------
    list
        A flat list
    """
    return [item for sublist in l for item in sublist]


def unique_by(l, fn):
    """Return a filtered list such that a given function is unique
    for each element of the list.

    Parameters
    ----------
    l : list
        A list of arbitrary objects.
    fn : function
        Must return a hashable object.

    Returns
    -------
    list
        A list of unique elements according to `fn`.
    """
    by_key = OrderedDict(zip([fn(_) for _ in l], l))
    return list(by_key.values())


def chunks(l, n):
    """Yield successive n-sized chunks from l.

    https://stackoverflow.com/a/312464

    Parameters
    ----------
    l : iterable
        The list or list-like object to be split into chunks.
    n : int
        The size of the chunks to be generated.

    Yields
    ------
    iterable
        Consecutive slices of l of size n.
    """
    for i in _range(0, len(l), n):
        yield l[i:i + n]


def equal_list_of_dicts(obj1, obj2, exclude=[]):
    """Check whether two lists of dictionaries are equal, independent of the
    order within the list.

    Parameters
    ----------
    obj1 : list of dict
        First object to compare
    obj2 : list of dict
        Second object to compare
    exclude : list of str, optional
        A list of keys that are to be excluded when deciding whether the lists
        of dictionaries are equal (default: []).

    Returns
    -------
    bool
        True if the two lists contain the same dictionaries, False otherwise.
    """
    for key in exclude:
        for i in obj1:
            del i[key]
        for i in obj2:
            del i[key]
    serial1 = [json.dumps(_, sort_keys=True) for _ in obj1]
    serial2 = [json.dumps(_, sort_keys=True) for _ in obj2]

    return set(serial1) == set(serial2)


def select(objects, unlist=True, first=False, **kwargs):
    """Returns a subset of `objects` that matches a range of criteria.

    Parameters
    ----------
    objects : list of dict, dict of dict
        The collection of objects to filter based on keys.
    first: bool, optional
        If True, return first entry only (default: False).
    unlist : bool, optional
        If True and the result has length 1 and objects is a list, return the
        object directly, rather than the list (default: True).
    kwargs : dict
        Criteria as key-value pairs that must be true for each item in
        `objects`.

    Returns
    -------
    list
        A list of all items in `objects` that match the specified criteria.

    Examples
    --------
    >>> select([{'a': 1, 'b': 2}, {'a': 2, 'b': 2}, {'a': 1, 'b': 1}], a=1)
    [{'a': 1, 'b': 2}, {'a': 1, 'b': 1}]
    """
    filtered_objects = objects
    if type(objects) is list:
        for key, val in kwargs.items():
            filtered_objects = [obj for obj in filtered_objects
                                if key in obj and obj[key] == val]
    elif type(objects) is dict:
        for key, val in kwargs.items():
            filtered_objects = {obj_key: obj for obj_key, obj
                                in filtered_objects.items()
                                if key in obj and obj[key] == val}
    if first:
        if type(filtered_objects) is list:
            return filtered_objects[0]
        elif type(filtered_objects) is dict:
            return filtered_objects[list(filtered_objects.keys())[0]]
    elif unlist and len(filtered_objects) == 1 and \
            type(filtered_objects) is list:
        return filtered_objects[0]
    else:
        return filtered_objects


# -----------------------------------------------------------------------------
# File system operations
# -----------------------------------------------------------------------------
def ls(directory, extensions=['.zip', '.tgz', '.nc'], subset=None, path=True):
    """Recursively list the files in a directory.

    Parameters
    ----------
    directory : str
        The path of the directory in which to search for files.
    extensions : list of str, optional
        A list of file extensions to include (default: ['.zip', '.tgz', '.nc'])
    subset : list of str, optional
        Restrict the returned list to filenames appearing in `subset`.
    path : bool, optional
        If True, return the full path of each file. Otherwise return only the
        filenames (default: True).

    Returns
    -------
    list of str
        A list of all files with matching extensions in the directory.
    """
    def _in_subset(f):
        if subset is None and os.path.splitext(f)[1].lower() in extensions:
            return True
        elif subset is not None:
            for s in subset:
                if f.startswith(s):
                    return True
            return False
        else:
            return False

    all_files = []
    for (root, subdirs, files) in os.walk(directory):
        for f in files:
            if _in_subset(f):
                if path:
                    all_files.append(os.path.join(root, f))
                else:
                    all_files.append(f)
    return all_files



from esahub import scihub, utils, checksum, check, main
import unittest
import contextlib
import logging
import re
import datetime as DT
import pytz
import os
import sys
import subprocess
from shapely.wkt import loads as wkt_loads
from esahub.tests import config as test_config
from esahub import config

logger = logging.getLogger('esahub')

PY2 = sys.version_info < (3, 0)
SMALL_SIZE_QUERY = 'size: ???.* KB'

if hasattr(unittest.TestCase, 'subTest'):
    class TestCase(unittest.TestCase):
        pass
else:
    class TestCase(unittest.TestCase):
        @contextlib.contextmanager
        def subTest(self, msg='', **params):
            """Mock subTest method so no exception is raised under Python2."""
            utils.eprint('subTest:', msg, params)
            yield
            return


# -----------------------------------------------------------------------------
# TEST SETUP
# -----------------------------------------------------------------------------
def setUpModule():
    test_config.set_test_config()
    test_config.prepare()


def tearDownModule():
    test_config.cleanup()


# -----------------------------------------------------------------------------
# SCIHUB
# -----------------------------------------------------------------------------
class ScihubTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    # def setUp(self):

    def test_servers(self):
        for name in scihub._get_available_servers():
            cfg = config.CONFIG['SERVERS'][name]
            with self.subTest(server_name=name):
                url = '{}/search?q=*:*'.format(cfg['host'])
                response = scihub.get_response(url)
                #
                # Assert that the HTML response has status code 200 (OK)
                #
                self.assertEqual(response.status, 200)

    def test__generate_next_url(self):
        # _generate_next_url(url, total=None)
        pass

    def test__parse_page(self):
        # _parse_page(url, first=False)
        pass

    def test__get_file_list_from_url(self):
        # _get_file_list_from_url(url, limit=None)
        pass

    def test__callback(self):
        # _callback(result)
        pass

    def test__build_query(self):
        # _build_query(query={})
        pass

    def test__build_url(self):
        # _build_url(query, server)
        pass

    def test__download(self):
        # _download(url, destination, quiet=None, queue=None)
        pass

    def test__get_file_list_wrapper(self):
        # _get_file_list_wrapper(url)
        pass

    def test__ping_single(self):
        # _ping_single(servername)
        pass

    def test__auto_detect_server_from_query(self):
        queries = [
            # (query, server)
            ({'mission': 'Sentinel-1'},
             config.CONFIG['SATELLITES']['S1A']['source']),
            ({'mission': 'Sentinel-2'},
             config.CONFIG['SATELLITES']['S2A']['source']),
            ({'mission': 'Sentinel-3'},
             config.CONFIG['SATELLITES']['S3A']['source']),
            ({'satellite': 'S1A'},
             config.CONFIG['SATELLITES']['S1A']['source']),
            ({'satellite': 'S3A'},
             config.CONFIG['SATELLITES']['S3A']['source']),
            ({'satellite': 'S2B'},
             config.CONFIG['SATELLITES']['S2B']['source']),
            ({'identifier': "S1A_IW_OCN__2SDV_20160924T181320_"
                            "20160924T181345_013198_014FDF_6692.zip"},
             config.CONFIG['SATELLITES']['S1A']['source'])
        ]
        for query, server in queries:
            with self.subTest(query=query):
                self.assertEqual(
                    scihub._auto_detect_server_from_query(query), server
                )

    def test__uuid_from_identifier(self):
        products = scihub.search({}, limit=1)
        for product in products:
            with self.subTest(product=product):
                self.assertEqual(
                    scihub.block(scihub._uuid_from_identifier,
                                 product['title']),
                    product['uuid']
                )

    # def test__download_url_from_identifier(self):
    #     # _download_url_from_identifier(identifier)
    #     pass
    # def test__checksum_url_from_identifier(self):
    #     # _checksum_url_from_identifier(identifier)
    #     pass
    # def test__preview_url_from_identifier(self):
    #     # _preview_url_from_identifier(identifier)
    #     pass
    # def test__download_url_from_uuid(self):
    #     # _download_url_from_uuid(uuid, host=None)
    #     pass
    # def test__checksum_url_from_uuid(self):
    #     # _checksum_url_from_uuid(uuid, host=None)
    #     pass
    # def test__preview_url_from_uuid(self):
    #     # _preview_url_from_uuid(uuid, host=None)
    #     pass

    def test_get_response(self):
        for name in scihub._get_available_servers():
            with self.subTest(server_name=name):
                response = scihub.get_response(
                    scihub._build_url({'query': '*:*'}, name)
                )
                self.assertEqual(response.status, 200)

    def test_md5_from_file(self):
        for f in utils.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the md5 sum computed from the local file is equal
                # to the md5 sum obtained from the remote server.
                #
                try:
                    remote_md5 = scihub.md5(f)
                    self.assertEqual(
                        checksum.md5(f), remote_md5
                    )
                except Exception as e:
                    self.fail('Remote MD5 could not be obtained: {}'.format(e))

    def test_exists_true(self):
        existing = scihub.search({}, limit=1)
        for e in existing:
            with self.subTest(product=e['filename']):
                self.assertTrue(scihub.exists(e['filename']))

    def test_exists_false(self):
        not_existing = 'this_is_not_on_scihub'
        self.assertFalse(scihub.exists(not_existing))


# -----------------------------------------------------------------------------
# SCIHUB SEARCH
# -----------------------------------------------------------------------------
class ScihubSearchTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    def test_query_entries(self):
        query = {'mission': 'Sentinel-3'}
        server = scihub._auto_detect_server_from_query(query,
                                                       available_only=True)[0]
        url = scihub._build_url(query, server)
        html = scihub.resolve(url)
        #
        # Assert that the number of entries found on the page matches the
        # number of entries requested per page.
        #
        self.assertEqual(html.count('<entry>'),
                         config.CONFIG['GENERAL']['ENTRIES'])

    def test_orbit_query(self):
        for search_str, orbit in [
            ('ASC', 'ASCENDING'),
            ('DESC', 'DESCENDING')
        ]:
            query = {'orbit': search_str}
            result = scihub.search(query, limit=20)
            for prod in result:
                self.assertEqual(prod['orbit_direction'], orbit)

    def test_id_query(self):
        prod = scihub.search({}, limit=5)[-1]
        query = {'id': prod['title']}
        result = scihub.search(query)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], prod)

    def test_queries(self):
        queries = [
            # (name, query)
            ('S3', {'mission': 'Sentinel-3'}),
        ]
        for name, q in queries:
            with self.subTest(name=name):
                server = scihub._auto_detect_server_from_query(
                    q, available_only=True)[0]
                url = scihub._build_url(q, server=server)
                response = scihub.get_response(url)
                #
                # Assert that queries for each mission return a
                # status code 200 (OK)
                #
                self.assertEqual(response.status, 200)

        with self.subTest('count entries'):
            q = {'mission': 'Sentinel-3'}
            server = scihub._auto_detect_server_from_query(
                q, available_only=True)[0]
            url = scihub._build_url(q, server=server)
            html = scihub.resolve(url)
            #
            # Assert that the number of entries found on the page matches the
            # number of entries requested per page.
            #
            self.assertEqual(html.count('<entry>'),
                             config.CONFIG['GENERAL']['ENTRIES'])

    def test_temporal_queries(self):
        with self.subTest('yesterday'):
            file_list = scihub.search({'mission': 'Sentinel-3',
                                       'time': 'yesterday'},
                                      limit=200)
            yesterday = DT.datetime.now(pytz.utc)-DT.timedelta(1)
            today = DT.datetime.now(pytz.utc)
            start = DT.datetime(yesterday.year, yesterday.month, yesterday.day,
                                tzinfo=pytz.utc)
            end = DT.datetime(today.year, today.month, today.day,
                              tzinfo=pytz.utc)
            for f in file_list:
                #
                # Assert that the ingestiondate of each entry was yesterday.
                #
                self.assertGreaterEqual(f['ingestiondate'], start)
                self.assertLessEqual(f['ingestiondate'], end)

        with self.subTest('today'):
            file_list = scihub.search({'mission': 'Sentinel-3',
                                       'time': 'today'},
                                      limit=200)
            today = DT.datetime.now(pytz.utc)
            start = DT.datetime(today.year, today.month, today.day,
                                tzinfo=pytz.utc)
            for f in file_list:
                #
                # Assert that the ingestiondate of each entry is today.
                #
                self.assertGreaterEqual(f['ingestiondate'], start)

    #
    # NOTE: This test presently fails because apparantly,
    # SciHub's `intersects` parameter does not work reliably.
    #
    def test_spatial_queries(self):
        loc, ref_coords = next(iter(config.CONFIG['LOCATIONS'].items()))
        with self.subTest(location=loc):
            file_list = scihub.search(
                {'location': [loc], 'time': 'to 2017-09-01T00:00:00Z'},
                server='S3', limit=20)
            for f in file_list:
                with self.subTest(product=f['filename']):
                    #
                    # Assert that the products indeed intersect the
                    # requested location.
                    #
                    distance = wkt_loads(f['coords']).distance(
                        wkt_loads(ref_coords))
                    utils.eprint('Distance: {}'.format(distance))
                    self.assertLessEqual(distance, 0.5)

    def test_get_file_list(self):
        q = {'mission': 'Sentinel-3'}
        limit = 107
        file_list = scihub.search(q, limit=limit)
        #
        # Assert that only `limit` entries are returned.
        #
        self.assertEqual(limit, len(file_list))
        for f in file_list:
            #
            # Assert that each entry contains the attributes `url`, `uuid` and
            # `filename`.
            #
            self.assertIn('url', f)
            self.assertIn('uuid', f)
            self.assertIn('filename', f)


# -----------------------------------------------------------------------------
# SCIHUB DOWNLOAD
# -----------------------------------------------------------------------------
class ScihubDownloadTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    def setUp(self):
        test_config.clear_test_data()

    def tearDown(self):
        test_config.clear_test_data()

    def test_download(self):
        file_list = scihub.search({'query': SMALL_SIZE_QUERY}, limit=1)
        for f in file_list:
            with self.subTest(url=f['url']):
                result = scihub.download(f)
                #
                # Assert that the download didn't fail and that
                # the returned file path exists.
                #
                self.assertNotEqual(result, False)
                self.assertTrue(os.path.isfile(result))

    def test_download_many(self):
        file_list = scihub.search({'query': SMALL_SIZE_QUERY},
                                  limit=2)
        scihub.download(file_list)
        #
        # Assert that all downloads were successful.
        #
        local_files = utils.ls(config.CONFIG['GENERAL']['DATA_DIR'])
        local_files_identifiers = [os.path.splitext(os.path.split(_)[1])[0]
                                   for _ in local_files]
        for f in file_list:
            self.assertIn(f['filename'], local_files_identifiers)
        for f in local_files:
            with self.subTest(file=f):
                _, healthy, msg = check.check_file(f, mode='file')
                utils.eprint(msg)
                self.assertTrue(healthy)

    def test_redownload(self):
        test_config.copy_corrupt_data()
        local_files = utils.ls(config.CONFIG['GENERAL']['DATA_DIR'])
        scihub.redownload(local_files)
        new_local_files = utils.ls(config.CONFIG['GENERAL']['DATA_DIR'])
        self.assertEqual(set(local_files), set(new_local_files))
        for f in local_files:
            with self.subTest(file=f):
                _, healthy, msg = check.check_file(f, mode='file')
                utils.eprint(msg)
                self.assertTrue(healthy)


# -----------------------------------------------------------------------------
# CHECK
# -----------------------------------------------------------------------------
class CheckTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    def setUp(self):
        test_config.copy_test_data()

    def tearDown(self):
        test_config.clear_test_data()

    def test_check_file_md5_healthy(self):
        for f in utils.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the files check out in `md5` mode.
                #
                try:
                    file_path, healthy, message = \
                        check.check_file(f, mode='md5')
                    self.assertTrue(healthy)
                except Exception as e:
                    self.fail('File check failed: {}'.format(e))

    def test_check_file_zip_healthy(self):
        for f in utils.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the files check out in `file` mode.
                #
                try:
                    file_path, healthy, message = \
                        check.check_file(f, mode='file')
                    self.assertTrue(healthy)
                except Exception as e:
                    self.fail('File check failed: {}'.format(e))

    def test_check_file_md5_corrupt(self):
        test_config.clear_test_data()
        test_config.copy_corrupt_data()
        for f in utils.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the files are detected as corrupt in `md5` mode.
                #
                try:
                    file_path, healthy, message = \
                        check.check_file(f, mode='md5')
                    self.assertFalse(healthy)
                except Exception as e:
                    self.fail('File check failed: {}'.format(e))

    def test_check_file_zip_corrupt(self):
        test_config.clear_test_data()
        test_config.copy_corrupt_data()
        for f in utils.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the files are detected as corrupt in `file` mode.
                #
                try:
                    file_path, healthy, message = \
                        check.check_file(f, mode='file')
                    self.assertFalse(healthy)
                except Exception as e:
                    self.fail('File check failed: {}'.format(e))


# -----------------------------------------------------------------------------
# CHECKSUM
# -----------------------------------------------------------------------------
class ChecksumTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    def setUp(self):
        test_config.copy_test_data()

    def tearDown(self):
        test_config.clear_test_data()

    def test_md5(self):
        for f in utils.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the md5 checksum returned by checksum.md5() is
                # equal to the md5 sum returned by bash md5 or md5sum tool.
                #
                for exe in ['md5', 'md5sum']:
                    if utils._which(exe) is not None:
                        bash_output = subprocess.check_output([exe, f])
                        if not PY2:
                            bash_output = bash_output.decode()
                        bash_md5 = re.search('[a-zA-Z0-9]{32}',
                                             bash_output).group()
                        break
                self.assertEqual(
                    checksum.md5(f), bash_md5
                )

    def test_etag_small_files(self):
        for f in utils.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the computed etag is equal to the md5
                # checksum for files smaller than the chunksize.
                #
                size_mb = max(10, int(os.path.getsize(f) / 1024**2))
                self.assertEqual(
                    checksum.md5(f), checksum.etag(f, chunksize=2 * size_mb)
                )

    # def test_etag_large_files(self):
    #     pass


# -----------------------------------------------------------------------------
# MAIN
# -----------------------------------------------------------------------------
class MainTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()
        cls.check_mode = config.CONFIG['GENERAL']['CHECK_MODE']
        config.CONFIG['GENERAL']['CHECK_MODE'] = 'file'

    @classmethod
    def tearDownClass(cls):
        config.CONFIG['GENERAL']['CHECK_MODE'] = cls.check_mode

    def setUp(self):
        test_config.copy_test_data()

    def tearDown(self):
        test_config.clear_all()

    def test_ls(self):
        q = {'time': 'today', 'satellite': 'S3A',
             'location': ['Ireland_Mace_Head']}
        files = scihub.search(q)
        result = main.ls(q)
        self.assertEqual(len(result), len(files))

    def test_get(self):
        test_config.clear_test_data()
        q = {'satellite': 'S3A', 'query': SMALL_SIZE_QUERY}
        files = scihub.search(q, limit=2)
        main.get(q, limit=2)

        for f in files:
            ext = '.zip'
            with self.subTest(product=f['filename']):
                self.assertTrue(
                    os.path.isfile(os.path.join(
                        config.CONFIG['GENERAL']['DATA_DIR'],
                        f['filename']) + ext)
                )

    def test_doctor(self):
        test_config.copy_corrupt_data()
        corrupt_files = utils.ls(test_config.TEST_DATA_DIR_CORRUPT,
                                 path=False)
        # healthy_files = utils.ls(test_config.TEST_DATA_DIR_ORIGINAL,
        #                          path=False)
        result = main.doctor()
        bad_files = [os.path.split(status[0])[1]
                     for status in result if status[1] is False]

        #
        # Assert that the number of healthy/corrupt files detected are correct
        #
        self.assertEqual(len(bad_files), len(corrupt_files))
        for corrupt_file in corrupt_files:
            #
            # Assert that each corrupt file has been registered.
            #
            self.assertIn(corrupt_file, bad_files)

    def test_doctor_delete(self):
        test_config.copy_corrupt_data()
        corrupt_files = utils.ls(test_config.TEST_DATA_DIR_CORRUPT,
                                 path=False)
        healthy_files = utils.ls(test_config.TEST_DATA_DIR_ORIGINAL,
                                 path=False)
        main.doctor(delete=True)
        #
        # Assert that the corrupt files have been deleted.
        #
        for f in corrupt_files:
            self.assertFalse(os.path.isfile(os.path.join(
                    config.CONFIG['GENERAL']['DATA_DIR'], f)))

        #
        # Assert that the healthy files have not been deleted.
        #
        for f in healthy_files:
            self.assertTrue(os.path.isfile(os.path.join(
                    config.CONFIG['GENERAL']['DATA_DIR'], f)))

    def test_doctor_repair(self):
        test_config.copy_corrupt_data()
        corrupt_files = utils.ls(test_config.TEST_DATA_DIR_CORRUPT,
                                 path=False)
        # healthy_files = utils.ls(test_config.TEST_DATA_DIR_ORIGINAL,
        #                            path=False)
        main.doctor(repair=True)

        for f in corrupt_files:
            repaired_f = os.path.join(config.CONFIG['GENERAL']['DATA_DIR'], f)
            with self.subTest(file=repaired_f):
                #
                # Assert that each corrupt file has been repaired.
                #
                _, healthy, msg = check.check_file(repaired_f, mode='file')
                utils.eprint(msg)
                self.assertTrue(healthy)


# -----------------------------------------------------------------------------
# utils
# -----------------------------------------------------------------------------

class UtilsTestCase(TestCase):
    def test_parse_datetime(self):
        _dt = DT.datetime
        dates = [
            ('Sep 5, 2016', (_dt(2016, 9, 5, 0, 0, 0),
                             _dt(2016, 9, 6, 0, 0, 0))),
            ('5 Sep 2016', (_dt(2016, 9, 5, 0, 0, 0),
                            _dt(2016, 9, 6, 0, 0, 0))),
            ('06/1998', (_dt(1998, 6, 1, 0, 0, 0),
                         _dt(1998, 7, 1, 0, 0, 0))),
            ('Jan 2018 to Oct 2018', (_dt(2018, 1, 1, 0, 0, 0),
                                      _dt(2018, 11, 1, 0, 0, 0))),
            ('1 Jan 2018 to 30 Sep 2018', (_dt(2018, 1, 1, 0, 0, 0),
                                           _dt(2018, 10, 1, 0, 0, 0))),
            ('12/2017', (_dt(2017, 12, 1, 0, 0, 0),
                         _dt(2018, 1, 1, 0, 0, 0))),
            ('2017/12', (_dt(2017, 12, 1, 0, 0, 0),
                         _dt(2018, 1, 1, 0, 0, 0))),
            ('2017/12 to 2018/12', (_dt(2017, 12, 1, 0, 0, 0),
                                    _dt(2019, 1, 1, 0, 0, 0))),
            ('Jan 1, 2017, Jan 1, 2018', (_dt(2017, 1, 1, 0, 0, 0),
                                          _dt(2018, 1, 2, 0, 0, 0))),
            ('to Jan 2018', (None, _dt(2018, 2, 1, 0, 0, 0))),
            ('2015 -', (_dt(2015, 1, 1, 0, 0, 0), None)),
            ('to 2017-09-01T00:00:00', (None, _dt(2017, 9, 1, 0, 0, 0)))
        ]
        for date_str, date_obj in dates:
            with self.subTest(date_str=date_str):
                self.assertEqual(
                    utils.parse_datetime(date_str),
                    date_obj
                )



import os
import aiohttp
import asyncio
import lxml.etree as ET
from datetime import datetime, timedelta
import pytz
import re
from .config import CONFIG
from . import utils, geo, checksum, tty
from urllib.parse import urlparse, parse_qs, urlencode
from collections import OrderedDict
import hashlib
import logging
logger = logging.getLogger('esahub')
logger.disabled = True


CHUNK = 64 * 1024
PREFIXES = {
    'os': 'http://a9.com/-/spec/opensearch/1.1/',
    'opensearch': 'http://a9.com/-/spec/opensearch/1.1/',
    'doc': 'http://www.w3.org/2005/Atom',
    'gml': 'http://www.opengis.net/gml'
}
DOWNLOAD_SUFFIX = '.download'
DOWNLOAD_URL_PATTERN = \
    "{host}/odata/v1/Products('{uuid}')/$value"
CHECKSUM_URL_PATTERN = \
    "{host}/odata/v1/Products('{uuid}')/Checksum/Value/$value"
PREVIEW_URL_PATTERN = \
    "{host}/odata/v1/Products('{uuid}')/Products('Quicklook')/$value"
DATETIME_FMT = '%Y-%m-%dT%H:%M:%S.000Z'


# -----------------------------------------------------------------------------
# HTTP CONNECTION METHODS
# This is all the async stuff
# -----------------------------------------------------------------------------
class NotFoundError(Exception):
    pass


class SessionManager():
    """
    Manage active HTTP sessions.

    Keep different sessions for queries and downloads, because concurrent
    downloads are limited, while queries are not.
    """
    def __init__(self, concurrent=None):
        self._concurrent = concurrent
        self._sessions = {}

    def __getitem__(self, server):
        if server not in self._sessions:
            cfg = CONFIG['SERVERS'][server]
            auth = aiohttp.BasicAuth(login=cfg['user'],
                                     password=cfg['password'])
            if self._concurrent is None:
                concurrent = cfg['downloads']
            else:
                concurrent = self._concurrent
            connector = aiohttp.TCPConnector(limit=concurrent)
            self._sessions[server] = aiohttp.ClientSession(
                auth=auth, connector=connector)
        return self._sessions[server]

    def __del__(self):
        # Close all active sessions
        loop = asyncio.get_event_loop()
        closer_tasks = []
        for server, session in self._sessions.items():
            closer_tasks.append(session.close())
        if len(closer_tasks) > 0:
            loop.run_until_complete(asyncio.wait(closer_tasks))


QUERY = SessionManager(concurrent=CONFIG['GENERAL']['N_SCIHUB_QUERIES'])
DOWNLOAD = SessionManager()


def block(fn, *args, **kwargs):
    """Run an async function and block."""
    task = fn(*args, **kwargs)
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(task)
    return result


def get_response(url):
    async def _resp(url):
        server = _get_server_from_url(url)
        async with QUERY[server].get(url) as resp:
            return resp
    return block(_resp, url)


def resolve(url, server=None):
    return block(_resolve, url, server=server)


async def _resolve(url, server=None):
    if server is None:
        server = _get_server_from_url(url)

    async with QUERY[server].get(url) as response:
        return await response.text()


async def get_total_results(url):
    response = await _resolve(url)
    xml = response.encode('utf-8')
    root = ET.fromstring(xml)
    try:
        total_results = int(root.find('os:totalResults', PREFIXES).text)
    except TypeError:
        total_results = 0
    except AttributeError as e:
        raise AttributeError("Could not extract total results from URL "
                             "{}: {}".format(url, e))
    return total_results


async def _ping_single(server):
    cfg = CONFIG['SERVERS'][server]
    url = '{host}/search?q=*:*'.format(host=cfg['host'])
    async with QUERY[server].get(url) as response:
        return (server, response.status)


async def _download(url, destination, return_md5=False, cont=True):
    """Downloads a file from the remote server into the specified destination.

    Parameters
    ----------
    url : str
        The source URL.
    destination : str
        The local target file path.
    cont : bool, optional
        Continue partial downloads (default: True).

    Returns
    -------
    bool
        True if successful, False otherwise.
    """
    #
    # Create directory.
    #
    path, file_name = os.path.split(destination)
    os.makedirs(path, exist_ok=True)
    pbar_key = file_name.rstrip(DOWNLOAD_SUFFIX)

    if return_md5:
        hash_md5 = hashlib.md5()

    headers = {}
    if cont and os.path.isfile(destination):
        local_size = os.path.getsize(destination)
        headers['Range'] = 'bytes={}-'.format(local_size)
        with open(destination, 'rb') as f:
            hash_md5.update(f.read())
        tty.screen.status(progress=local_size)
    else:
        local_size = 0

    server = _get_server_from_url(url)
    async with DOWNLOAD[server].get(url, timeout=None, headers=headers) \
            as response:
        size = int(response.headers['Content-Length'])
        # Create progress bar only now:
        pbar = tty.screen[pbar_key]
        pbar.n = local_size
        pbar.total = size + local_size
        pbar.refresh()

        mode = 'ab' if cont else 'wb'
        with open(destination, mode) as f:
            async for data in response.content.iter_chunked(CHUNK):
                if return_md5:
                    hash_md5.update(data)
                f.write(data)
                progress = len(data)
                tty.screen.status(progress=progress)
                pbar.update(len(data))

    # if pbar is not None:
    #     pbar.close()

    if return_md5:
        return (True, hash_md5.hexdigest().lower())
    else:
        return True


# -----------------------------------------------------------------------------
# XML PARSING
# -----------------------------------------------------------------------------
def parse_page(xml):
    file_list = []

    try:
        root = ET.fromstring(xml)
        # try:
        #     total_results = int(root.find('os:totalResults', PREFIXES).text)
        # except TypeError:
        #     total_results = 0

        for entry in root.findall('doc:entry', PREFIXES):

            filename = entry.find("./doc:str[@name='identifier']",
                                  PREFIXES).text
            ingestiondate = utils.to_date(
                entry.find("./doc:date[@name='ingestiondate']", PREFIXES).text,
                output='date')
            try:
                footprint_tag = entry.find("./doc:str[@name='gmlfootprint']",
                                           PREFIXES).text
                match = re.search('<gml:coordinates>(.*)</gml:coordinates>',
                                  footprint_tag)
                coords = geo.gml_to_polygon(match.groups()[0])
            except AttributeError:
                coords = None

            filesize = utils.h2b(entry.find("./doc:str[@name='size']",
                                            PREFIXES).text)
            preview_url = entry.find("./doc:link[@rel='icon']",
                                     PREFIXES).attrib['href']
            try:
                rel_orbit = int(entry.find(
                    "doc:int[@name='relativeorbitnumber']", PREFIXES).text)
            except AttributeError:
                rel_orbit = None

            try:
                orbit_dir = entry.find("doc:str[@name='orbitdirection']",
                                       PREFIXES).text.upper()
            except AttributeError:
                orbit_dir = None

            file_dict = {
                'title': entry.find('doc:title', PREFIXES).text,
                'url': entry.find('doc:link', PREFIXES).attrib['href'],
                'preview': preview_url,
                'uuid': entry.find('doc:id', PREFIXES).text,
                'filename': filename,
                'size': filesize,
                'ingestiondate': ingestiondate,
                'coords': coords,
                'orbit_direction': orbit_dir,
                'rel_orbit': rel_orbit
            }
            file_dict['host'] = _get_host_from_url(file_dict['url'])
            file_list.append(file_dict)

    except ET.XMLSyntaxError:
        # not valid XML
        file_list = []

    return file_list


async def _files_from_url(url, verbose=False):
    xml = await _resolve(url)
    result = parse_page(xml.encode('utf-8'))
    if verbose:
        tty.screen.status(progress=len(result))
    return result


async def _get_file_list_from_url(url, limit=None, verbose=False):
    # Parse first page to get total number of results.
    total_results = await get_total_results(url)
    host = urlparse(url).netloc
    if verbose:
        tty.screen.status(desc='Querying {host}'.format(host=host),
                          total=total_results, mode='bar')

    if limit is None:
        total = total_results
    else:
        total = min(limit, total_results)

    urls = [url] + [u for u in _generate_next_url(url, total=total)]

    tasks = [_files_from_url(u, verbose=verbose) for u in urls]
    results = await asyncio.gather(*tasks)
    result = utils.flatten(results)

    return result


# -----------------------------------------------------------------------------
# QUERY BUILDING
# -----------------------------------------------------------------------------
def _parse_time_parameter(value):
    # Default ingestiontime query parameters
    start = '1970-01-01T00:00:00.000Z'
    end = 'NOW'
    DATE_FMT = '%Y-%m-%dT00:00:00.000Z'

    if value == 'today':
        start = datetime.strftime(datetime.now(pytz.utc), DATE_FMT)

    elif value == 'yesterday':
        start = datetime.strftime(datetime.now(pytz.utc)
                                  - timedelta(1), DATE_FMT)
        end = datetime.strftime(datetime.now(pytz.utc), DATE_FMT)

    elif value == 'midnight':
        end = datetime.strftime(datetime.now(pytz.utc), DATE_FMT)

    elif value == '24h':
        start = datetime.strftime(datetime.now(pytz.utc)
                                  - timedelta(1), DATETIME_FMT)
        end = datetime.strftime(datetime.now(pytz.utc), DATETIME_FMT)

    else:
        parsed = utils.parse_datetime(value)
        if parsed[0] is not None:
            start = datetime.strftime(parsed[0], DATETIME_FMT)
        if parsed[1] is not None:
            end = datetime.strftime(parsed[1], DATETIME_FMT)

    return start, end


def _build_query(query={}):
    """ Builds and returns the query URL given the command line input
    parameters.

    Parameters
    ----------
    query : list, optional
        A list of additional custom query elements submitted to SciHub. The
        queries will be concatenated with ampersands (&).
    """
    query_list = []
    sort_string = ''

    # Default ingestiontime query parameters
    start = '1970-01-01T00:00:00.000Z'
    end = 'NOW'
    # Geospatial query list
    geo_query_list = []

    for key, val in query.items():
        if key == 'mission':
            # Build mission selection query:
            query_list.append('platformname:{0}'.format(query['mission']))

        elif key == 'satellite':
            # Build satellite selection query:
            query_list.append('identifier:{0}*'.format(query['satellite']))

        elif key == 'time':
            start, end = _parse_time_parameter(val)

        elif key == 'geo':
            # Build geospatial query:
            if type(val) is not list:
                val = [val]
            for item in val:
                geo_query_list.append(
                        'footprint:"Intersects({0})"'.format(item))

        elif key == 'location':
            if type(val) is not list:
                val = [val]
            for loc in val:
                if loc in CONFIG['LOCATIONS']:
                    geo_query_list.append('footprint:"Intersects({0})"'.format(
                            CONFIG['LOCATIONS'][loc]))
                else:
                    logger.error(
                        '{0} {1}'.format(tty.error('Location not found:'), loc)
                    )

        elif key == 'type':
            query_list.append('producttype:{}'.format(query['type']))

        elif key == 'orbit':
            if val.upper() in ['ASC', 'ASCENDING']:
                orbit = 'ASCENDING'
            elif val.upper() in ['DESC', 'DESCENDING']:
                orbit = 'DESCENDING'
            else:
                raise ValueError("Invalid value for `orbit`: '{}'"
                                 .format(val))
            query_list.append('orbitdirection:{}'.format(orbit))

        elif key == 'id':
            query_list.append('identifier:{}'.format(query['id']))

        elif key == 'query':
            query_list.append('{}'.format(query['query']))

        elif key == 'sort':
            sort_string = '&orderby={} {}'.format(*query['sort'])

        # Not a special keyword. Pass directly to SciHub
        else:
            query_list.append('{}:{}'.format(key, val))

    # Build ingestiondate query:
    query_list.append('ingestiondate:[{0} TO {1}]'.format(start, end))

    # Build geospatial query:
    if len(geo_query_list):
        query_list.append(
            '({})'.format(' OR '.join(geo_query_list))
        )

    # Generate full query string
    if len(query_list) == 0:
        query_list.append('*:*')
    query_string = ' AND '.join(query_list)

    #
    # Example: 'https://scihub.copernicus.eu/s3/search?
    #           q=(footprint:"Intersects(POLYGON((-25.100 46.800,
    #           -5.250 46.800,-5.250 57.400,-25.100 57.400,-25.100 46.800)))")
    #           &rows=25&start=0'
    #
    query_url = 'q={q}&start=0&rows={rows}{sort}'.format(
        q=query_string,
        rows=CONFIG['GENERAL']['ENTRIES'],
        sort=sort_string
    )
    return query_url


def _build_url(query, server):
    if type(query) is dict:
        query_string = _build_query(query)
    else:
        query_string = query
    return '{url}/search?{q}'.format(
        url=CONFIG['SERVERS'][server]['host'], q=query_string
    )


# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS
# -----------------------------------------------------------------------------
def _generate_next_url(url, total=None):
    parsed_url = urlparse(url)
    q_params = parse_qs(parsed_url.query)
    q_params = {k: v[0] if len(v) == 1 else v for k, v in q_params.items()}
    if 'rows' in q_params:
        q_params['rows'] = int(q_params['rows'])
    else:
        q_params['rows'] = 10
    if 'start' in q_params:
        q_params['start'] = int(q_params['start'])
    else:
        q_params['start'] = 0

    if total is not None:
        last_start = total - q_params['rows']
        q_params['rows'] = min(q_params['rows'], total)

    while total is None or q_params['start'] < last_start:
        q_params['start'] += q_params['rows']
        parsed_url = parsed_url._replace(query=urlencode(q_params, doseq=True))
        yield parsed_url.geturl()


def _get_available_servers():
    servers = [server for server, conf in CONFIG['SERVERS'].items()
               if len(conf['user']) > 0 and len(conf['password']) > 0]
    return servers


def _get_server_from_url(url):
    if 'server' in CONFIG['GENERAL']['QUERY']:
        return CONFIG['GENERAL']['QUERY']['server']
    for servername, cfg in CONFIG['SERVERS'].items():
        if cfg['host'] in url:
            return servername

    raise Exception("Could not determine server for {url}!".format(url=url))


def _get_host_from_url(url):
    p = urlparse(url)
    host = '{}://{}/{}'.format(
        p.scheme, p.netloc, p.path.strip('/').split('/')[0])
    return host


def _auto_detect_server_from_query(query, available_only=False):
    servers = None

    if 'satellite' in query and query['satellite'] in CONFIG['SATELLITES']:
        servers = CONFIG['SATELLITES'][query['satellite']]['source']

    if 'identifier' in query:
        sat = query['identifier'][:3]
        if sat in CONFIG['SATELLITES']:
            servers = CONFIG['SATELLITES'][sat]['source']

    if 'mission' in query:
        sats = utils.select(CONFIG['SATELLITES'], platform=query['mission'])
        if len(sats) > 0:
            ll = [source for sat in sats.values() for source in sat['source']]
            servers = list(OrderedDict.fromkeys(ll))

    #
    # If the server couldn't be determined from the query, return a list of
    # all servers.
    #
    if servers is None:
        servers = list(CONFIG['SERVERS'].keys())

    if available_only:
        servers = [server for server in servers
                   if server in _get_available_servers()]

    return servers


async def _uuid_from_identifier(identifier):
    identifier = os.path.splitext(os.path.split(identifier)[1])[0]
    results = await _search({'identifier': identifier+'*'})
    if len(results) == 0:
        raise NotFoundError('Product not found: {}'.format(identifier))
    return results[0]['uuid']


async def _host_and_uuid_from_identifier(identifier):
    identifier = os.path.splitext(os.path.split(identifier)[1])[0]
    results = await _search({'identifier': identifier+'*'})
    if len(results) == 0:
        raise NotFoundError('Product not found: {}'.format(identifier))
    return (results[0]['host'], results[0]['uuid'])


def _host_from_uuid(uuid):
    #
    # IMPLEMENT THIS
    #
    return None


async def _download_url_from_identifier(identifier):
    host, uuid = await _host_and_uuid_from_identifier(identifier)
    return _download_url_from_uuid(uuid, host=host)


async def _checksum_url_from_identifier(identifier):
    host, uuid = await _host_and_uuid_from_identifier(identifier)
    return _checksum_url_from_uuid(uuid, host=host)


async def _preview_url_from_identifier(identifier):
    host, uuid = await _host_and_uuid_from_identifier(identifier)
    return _preview_url_from_uuid(uuid, host=host)


def _download_url_from_uuid(uuid, host=None):
    if host is None:
        host = _host_from_uuid(uuid)
    return DOWNLOAD_URL_PATTERN.format(host=host, uuid=uuid)


def _checksum_url_from_uuid(uuid, host=None):
    if host is None:
        host = _host_from_uuid(uuid)
    return CHECKSUM_URL_PATTERN.format(host=host, uuid=uuid)


def _preview_url_from_uuid(uuid, host=None):
    if host is None:
        host = _host_from_uuid(uuid)
    return PREVIEW_URL_PATTERN.format(host=host, uuid=uuid)


# -----------------------------------------------------------------------------
# PUBLIC API METHODS
# -----------------------------------------------------------------------------
# def ping(server=None):
#     results = {}
#     if server is not None:
#         servers = [server]
#     else:
#         servers = list(CONFIG['SERVERS'].keys())

#     pool = multiprocessing.Pool(processes=len(servers))
#     results = pool.map_async(_ping_single, servers)
#     # pool.close()
#     # pool.join()
#     # for servername,server in servers.items():
#     while not results.ready():
#         time.sleep(1)
#     return results.get()

def search(*args, **kwargs):
    return block(_search, *args, **kwargs)


async def _search(query={}, server='auto', limit=None, verbose=False,
                  **kwargs):
    """ Search SciHub for satellite products.
    Parameters
    ----------
    query : dict, optional
        Query parameters as dictionary. May contain key 'server' (see below).
    server : str, optional
        If server is 'all', search all saved SciHub servers.
        If server is 'auto', attempt to auto-detect the server from the query.
        Otherwise, use the specified server (as configured in CONFIG.SERVERS)
        (default: 'auto')
    limit : int, optional
        The maximum number of results to return.
    kwargs : dict, optional
        The query parameters can also be passed as keyword arguments.

    Returns
    -------
    list of dict
        A list of dictionaries representing the found search results.
    """
    #
    # Search in each server.
    #
    query.update(kwargs)
    if 'server' in query:
        server = query.pop('server')

    if server == 'all':
        servers = _get_available_servers()
    elif server == 'auto':
        servers = _auto_detect_server_from_query(query, available_only=True)
    else:
        servers = [server]

    if servers is None:
        servers = []
    query_string = _build_query(query)

    tasks = []
    for servername in servers:
        server = CONFIG['SERVERS'][servername]
        url = '{url}/search?{q}'.format(
            url=server['host'], q=query_string
        )
        logger.debug('Trying server {}: {}'.format(servername, url))
        tasks.append(
            _get_file_list_from_url(url, limit=limit, verbose=verbose)
        )
    results = await asyncio.gather(*tasks)
    results = utils.flatten(results)

    #
    # Delete duplicate results (if product is on multiple servers).
    # TODO: This should be done in the order of preference as
    # given by CONFIG['SERVERS'] !
    #
    unique = utils.unique_by(results, lambda x: x['filename'])
    if limit is not None:
        unique = unique[:limit]

    return unique


async def _md5(product=None, uuid=None):
    if product is not None:
        if type(product) is dict and 'uuid' in product and 'host' in product:
            md5_url = _checksum_url_from_uuid(product['uuid'],
                                              host=product['host'])
        elif type(product) is str:
            md5_url = await _checksum_url_from_identifier(product)
        else:
            return False

    elif uuid is not None:
        md5_url = _checksum_url_from_uuid(uuid)

    server = _get_server_from_url(md5_url)
    async with QUERY[server].get(md5_url) as response:
        result = await response.read()
    return result.decode().lower()

    # if PY2:
    #     return result.lower()
    # else:
    #     return result.decode().lower()


def md5(product=None, uuid=None):
    """Returns the md5 sum of the file stored on SciHub given the product name
    or uuid.

    Parameters
    ----------
    product : str, optional
        The product name. If given, `uuid` is ignored.
    uuid : str, optional
        The product uuid.

    Returns
    -------
    str
        The md5 checksum in lower case.
    """
    return block(_md5, product=product, uuid=uuid)


def exists(product):
    search_results = search({'identifier': product+'*'}, limit=1)
    return len(search_results) > 0


def download(product):
    cont = CONFIG['GENERAL']['CONTINUE']
    if isinstance(product, list):
        # Multiple downloads
        # tty.screen.status(total=len(product))
        tasks = [_single_download(p, return_md5=True, cont=cont)
                 for p in product]
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(asyncio.gather(*tasks))
    else:
        # Single download
        result = block(_single_download, product=product, cont=cont)

    return result


async def _single_download(product, return_md5=False, cont=True):
    """Download a satellite product.

    Checks for file existence and MD5 checksum.

    Parameters
    ----------
    product : str or dict
        The name of the product to be downloaded from SciHub.
        Alternatively, a dictionary representing a search result from SciHub.
    return_md5 : bool, optional
        Whether to compute and return the md5 hash sum (default: False).
    cont : bool, optional
        Continue partial downloads (default: True).

    Returns
    -------
    str
        The local file path if the download was successful OR the file already
        exists and passes the md5 checksum test. False otherwise.
    """
    if type(product) is dict:
        fdata = product
    else:
        fdata = await _search(
            {'identifier': os.path.splitext(product)[0]+'*'}
        )
        fdata = fdata[0]

    satellite = utils.get_satellite(fdata['filename'])
    ext = CONFIG['SATELLITES'][satellite]['ext']
    file_name = fdata['filename'] + ext

    pbar_key = file_name

    b_download = True
    b_file_okay = False
    complete = False

    full_file_path = os.path.join(CONFIG['GENERAL']['DATA_DIR'], file_name)
    download_path = full_file_path + DOWNLOAD_SUFFIX

    #
    # Check if file already exists in location:
    #
    if os.path.exists(download_path) and os.path.isfile(download_path):

        # Create progress bar
        # pbar = tty.screen[pbar_key]

        if not CONFIG['GENERAL']['CHECK_EXISTING']:
            #
            # File exists and won't be checked.
            #
            msg = '{} Skipping existing - MD5 not checked'.format(file_name)
            tty.screen[pbar_key] = (tty.warn('Skipping') + ': {name}',
                                    tty.NOBAR)
            logger.debug(msg)
            b_download = False

        else:
            #
            # File exists and will be checked for md5 consistency
            #
            local_md5 = checksum.md5(download_path)
            remote_md5 = await _md5(fdata)
            if local_md5 == remote_md5:
                file_size = os.path.getsize(download_path)
                msg = '{} Skipping download (MD5 okay)'.format(file_name)
                tty.screen[pbar_key] = (tty.success('Exists') + ': {name}',
                                        tty.NOBAR)
                tty.screen.status(progress=file_size)
                logger.debug(msg)
                b_download = False
                b_file_okay = True

            elif cont:
                # Continue download
                b_download = True

            else:
                msg = '{} MD5 wrong: redownloading ...'.format(file_name)
                # Don't create the progress bar just yet
                logger.debug(msg)

    if b_download:
        #
        # Retrials in case the MD5 hashsum fails
        #
        for i in range(CONFIG['GENERAL']['TRIALS']):
            complete = await _download(fdata['url'], download_path,
                                       return_md5=return_md5, cont=cont)
            if return_md5:
                complete, local_md5 = complete

            #
            # After download, check MD5 hashsum
            #
            if not complete:
                #
                # Download incomplete.
                #
                msg = '{} Download failed, trial {:d}/{:d}.'.format(
                        file_name, i+1, CONFIG['GENERAL']['TRIALS'])
                logger.debug(msg)
                tty.screen[pbar_key] = (tty.error('Failed') + ': {name}',
                                        tty.NOBAR)
            else:
                #
                # Download completed.
                #
                if not return_md5:
                    local_md5 = checksum.md5(download_path)

                remote_md5 = await _md5(fdata)
                if local_md5 != remote_md5:
                    #
                    # Download failed.
                    #
                    msg = '{} MD5 checksum failed, trial {:d}/{:d}.'.format(
                            file_name, i+1, CONFIG['GENERAL']['TRIALS'])
                    logger.debug(msg)
                    tty.screen[pbar_key] = (tty.error('Failed') + ': {name}',
                                            tty.NOBAR)
                else:
                    #
                    # Download completed and successful.
                    #
                    msg = '{} MD5 okay'.format(file_name)
                    logger.debug(msg)
                    tty.screen[pbar_key] = (tty.success('MD5 okay') +
                                            ': {name}', tty.NOBAR)
                    b_file_okay = True
                    break

        if not b_file_okay:
            msg = '{} Download failed.'.format(file_name)
            logger.warning(msg)
            tty.screen[pbar_key] = (tty.error('Failed') + ': {name}',
                                    tty.NOBAR)

    if CONFIG['GENERAL']['DOWNLOAD_PREVIEW']:
        full_preview_path = os.path.join(CONFIG['GENERAL']['DATA_DIR'],
                                         fdata['filename']+'.jpeg')
        #
        # If not yet done, download preview file
        #
        if not os.path.exists(full_preview_path) or \
                not os.path.isfile(full_preview_path):
            if not _download(fdata['preview'], full_preview_path):
                logger.info('  Preview not available.')

    if b_file_okay:
        #
        # File has been downloaded successfully OR already exists
        # --> Return the file path
        #
        os.rename(download_path, full_file_path)
        msg = 'Download successful: {}'.format(full_file_path)
        logger.debug(msg)
        tty.screen[pbar_key] = (tty.success('Successful') + ': {name}',
                                tty.NOBAR)
        if return_md5:
            return full_file_path, local_md5
        else:
            return full_file_path

    elif b_download and not complete:
        #
        # File download failed --> Return FALSE
        #
        msg = 'Download failed: {}'.format(file_name)
        logger.error(msg)
        tty.screen[pbar_key] = (tty.error('Failed') + ': {name}',
                                tty.NOBAR)
        return False

    else:
        #
        # File download was skipped --> Return FALSE
        #
        return False


async def _get_remote_files_per_satellite(files, satellite):
        product_names = [os.path.splitext(os.path.split(fpath)[1])[0]
                         for fpath in files]
        chunksize = 10
        queries = []
        for products in utils.chunks(product_names, chunksize):
            query_string = '({})'.format(
                    ' OR '.join(['identifier:{}'.format(product)
                                 for product in products])
            )
            queries.append({'query': query_string, 'satellite': satellite})
        tasks = [_search(q) for q in queries]
        result = await asyncio.gather(*tasks)
        return utils.flatten(result)


def _get_remote_files(files):
    tasks = []
    for sat in CONFIG['SATELLITES']:
        sat_files = [f for f in files
                     if os.path.split(f)[1].startswith(sat)]
        tasks.append(_get_remote_files_per_satellite(sat_files, sat))
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(asyncio.gather(*tasks))
    return utils.flatten(results)


def redownload(local_file_list):
    """Redownload a list of corrupt files.

    This method will automatically determine the correct source of each file
    from the file name. It will then attempt to redownload the files.

    Parameters
    ----------
    local_file_list : list of str
        A list of local files to be redownloaded from the server.
    """
    remote_files = _get_remote_files(local_file_list)
    logger.info('DOWNLOADING {}'.format(len(remote_files)))
    download(remote_files)



""" This module
"""
import sys
import binascii
import hashlib

PY2 = sys.version_info < (3, 0)


def md5(filename):
    """Compute the MD5 hashsum of a local file.

    Parameters
    ----------
    filename : str

    Returns
    -------
    str
        The md5 checksum in lower case.
    """
    hash_md5 = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest().lower()


def etag(filename, chunksize, system='swift'):
    """Compute the SWIFT etag of a local file.

    If the file is smaller than the chunksize, this is equivalent to the md5
    hashsum. Otherwise, it is the md5 of the concatenated md5 sums of the
    segments.

    Parameters
    ----------
    filename : str
    chunksize : int
        S3 chunksize in MB
    system : str
        One of 'swift' and 'wos'.

    Returns
    -------
    str
        The SWIFT/WOS etag in lower case.
    """
    oneMB = 1024**2
    segment_md5s = []

    with open(filename, 'rb') as f:
        #
        # Read file in 1MB chunks to avoid memory issues
        #
        chunk_counter = 0
        md5 = hashlib.md5()
        for chunk in iter(lambda: f.read(oneMB), b""):
            md5.update(chunk)
            chunk_counter += 1
            if chunk_counter % chunksize == 0:
                #
                # Append the md5 of this chunk and reset.
                #
                segment_md5s.append(md5.hexdigest())
                md5 = hashlib.md5()
        #
        # If there is a 'remainder' left, append the md5.
        #
        if chunk_counter % chunksize != 0:
            segment_md5s.append(md5.hexdigest())

    if len(segment_md5s) == 1:
        return segment_md5s[0].lower()

    if system.lower() == 'swift':
        if PY2:
            return hashlib.md5(''.join(segment_md5s)).hexdigest().lower()
        else:
            return hashlib.md5(''.join(segment_md5s).encode('utf-8')) \
                          .hexdigest().lower()
    elif system.lower() == 'wos':
        if PY2:
            return hashlib.md5(binascii.unhexlify(
                                ''.join(segment_md5s).encode('utf-8'))) \
                .hexdigest().lower() + '-{}'.format(len(segment_md5s))
        else:
            return hashlib.md5(binascii.unhexlify(
                                ''.join(segment_md5s).encode('utf-8'))) \
                .hexdigest().lower() + '-{}'.format(len(segment_md5s))



from .config import CONFIG
import sys
import re
from tqdm import tqdm
from tqdm._utils import _environ_cols_wrapper


tqdm.monitor_interval = 1


def TERM_WIDTH():
    if sys.stdout.isatty():
        return _environ_cols_wrapper()(sys.stdout)
    else:
        return 0


BAR = "{desc} {percentage:3.0f}% |{bar}| " \
      "{n_fmt: >5}/{total_fmt: <5} " \
      "[{elapsed}, {rate_fmt: >8}{postfix}]"
# The explicit width is necessary because tqdm doesn't handle color codes
# correctly.
NOBAR = "{desc: <%d}" % TERM_WIDTH()
STATUS_RATE = "{desc} {n_fmt} [{elapsed}, {rate_fmt}]"
STATUS_BAR = "{desc} {percentage:3.0f}% |{bar}| [{elapsed}]"
STATUS_NOBAR = "{desc} [{elapsed}]"
DESC_LEN = 50
LONG_DESC_LEN = 80
FAKE_TOTAL = 9000000000000
RE_ANSI = re.compile(r"\x1b\[[;\d]*[A-Za-z]")


def shorten(text, maxlen=30):
    if len(text) <= maxlen:
        return text
    else:
        n_end = int((maxlen - 3) / 2)
        n_start = maxlen - 3 - n_end
        return '{}...{}'.format(text[:n_start], text[-n_end:])


def _format_desc(desc, key, long=False):
    # Strip away color control codes for length calculation:
    _stripped = RE_ANSI.sub('', desc)
    current_len = len(_stripped)
    if '{name}' in desc:
        current_len -= len('{name}')
    if long:
        key_len = LONG_DESC_LEN - current_len
    else:
        key_len = DESC_LEN - current_len
    short_file_name = shorten(key, key_len)
    return desc.format(name=short_file_name)


class Screen():
    _status = None
    _result = None
    _lines = {}

    def __init__(self, *args, **kwargs):
        pass

    def status(self, desc=None, mode=None, total=None, progress=None,
               reset=False, unit=None, scale=None):
        if mode is None:
            bar_format = STATUS_NOBAR
        elif mode == 'rate':
            bar_format = STATUS_RATE
        elif mode == 'bar':
            bar_format = STATUS_BAR
        elif mode == 'static':
            bar_format = STATUS_NOBAR
        else:
            bar_format = STATUS_NOBAR
        if self._status is None:
            #
            # Initialize new status bar.
            #
            if total is None:
                total = FAKE_TOTAL
            if progress is None:
                progress = 0
            if unit is None:
                unit = 'B'
            if scale is None:
                scale = True
            self._status = tqdm(total=total, desc=desc,
                                disable=CONFIG['GENERAL']['QUIET'],
                                bar_format=bar_format,
                                initial=progress,
                                unit=unit, unit_scale=scale,
                                mininterval=1, maxinterval=1,
                                position=0,
                                leave=True)

        else:
            #
            # Update existing status bar.
            #
            if desc is not None:
                self._status.set_description_str(desc)
            if mode is not None:
                self._status.bar_format = bar_format
            if total is not None:
                self._status.total = total
            if progress is not None:
                self._status.update(progress)
            if reset:
                self._status.n = 0
            if unit is not None:
                self._status.unit = unit
            if scale is not None:
                self._status.unit_scale = scale

            self._status.refresh()

    def result(self, text):
        # Should only be called after all progress bars have been created.
        if self._result is None:
            try:
                disable = CONFIG['GENERAL']['QUIET']
            except TypeError:
                disable = False
            self._result = tqdm(disable=disable,
                                total=FAKE_TOTAL,
                                desc=text,
                                bar_format=NOBAR,
                                position=len(self._lines) + 1,
                                leave=True)
        else:
            self._result.set_description_str(text)

    def __getitem__(self, key):
        if key not in self._lines:
            self._lines[key] = tqdm(
                desc=_format_desc('Downloading {name}', key),
                unit='B', unit_scale=True,
                bar_format=BAR,
                mininterval=0.3, maxinterval=1,
                disable=CONFIG['GENERAL']['QUIET'],
                position=len(self._lines) + 1,
                leave=True
            )

        return self._lines[key]

    def __setitem__(self, key, value):
        pbar = self.__getitem__(key)
        if isinstance(value, tuple):
            value, fmt = value
            pbar.bar_format = fmt
            if pbar.total is None:
                pbar.total = FAKE_TOTAL
        pbar.desc = _format_desc(value, key)
        pbar.refresh()

    def quit(self):
        # Close all progress bars
        if self._result is None:
            self.result('done.')
        if self._status is not None:
            del self._status
        keys = list(self._lines.keys())
        for key in keys:
            del self._lines[key]
        if not self._result.disable:
            # NOTE: The following line is required to put the cursor at the end
            # of the progress bars upon termination. No idea why! :-(
            self._result.moveto(1 - self._result.pos)
        del self._result

    def __del__(self):
        self.quit()


screen = Screen()


# TERMINAL OUTPUT FORMATTING
# -----------------------------------------------------------------------------
ANSI = dict(
    PURPLE='\033[95m',
    CYAN='\033[96m',
    DARKCYAN='\033[36m',
    BLUE='\033[94m',
    GREEN='\033[92m',
    YELLOW='\033[93m',
    RED='\033[91m',
    BOLD='\033[1m',
    UNDERLINE='\033[4m',
    END='\033[0m'
)


def error(message):
    if sys.stdout.isatty():
        return '{BOLD}{RED}{text}{END}'.format(text=message, **ANSI)
    else:
        return message


def warn(message):
    if sys.stdout.isatty():
        return '{BOLD}{YELLOW}{text}{END}'.format(text=message, **ANSI)
    else:
        return message


def success(message):
    if sys.stdout.isatty():
        return '{BOLD}{GREEN}{text}{END}'.format(text=message, **ANSI)
    else:
        return message



# coding=utf-8
""" Helper module for geospatial tasks.
"""
import shapely.ops
from shapely.wkt import loads as wkt_loads
from functools import partial
try:
    import pyproj
    PYPROJ_INSTALLED = True
except ImportError:
    PYPROJ_INSTALLED = False


def gml_to_polygon(footprint):
    """Converts a GML footprint into a WKT polygon string.

    Can handle the formats used by Sentinel-1, 2, and 3

    Parameters
    ----------
    footprint : str
        A GML footprint as retrieved from SciHub.

    Returns
    -------
    str
        The converted WKT polygon string.
    """
    footprint = footprint.replace('\n', '').strip()
    coords_poly = []
    #
    # Sentinel-1
    # (http://www.opengis.net/gml/srs/epsg.xml#4326)
    #
    if ',' in footprint:
        coords_gml = footprint.split()
        for coord_pair in coords_gml:
            lat, lon = [float(_) for _ in coord_pair.split(',')]
            if lon < -180.0:
                lon = -180.0
            if lon > 180.0:
                lon = 180.0
            if lat < -90.0:
                lat = -90.0
            if lat > 90.0:
                lat = 90.0
            coords_poly.append('{lon:.4f} {lat:.4f}'.format(lon=lon, lat=lat))

    #
    # Sentinel-3 and Sentinel-2
    # (http://www.opengis.net/def/crs/EPSG/0/4326)
    #
    else:
        coords_gml = footprint.split()
        for i in range(len(coords_gml)//2):
            lat = float(coords_gml[2*i])
            lon = float(coords_gml[2*i+1])
            if lon < -180.0:
                lon = -180.0
            if lon > 180.0:
                lon = 180.0
            if lat < -90.0:
                lat = -90.0
            if lat > 90.0:
                lat = 90.0
            coords_poly.append('{lon:.4f} {lat:.4f}'.format(lon=lon, lat=lat))

    #
    # Make sure the polygon is a closed line string.
    #
    if coords_poly[0] != coords_poly[-1]:
        coords_poly.append(coords_poly[0])

    wkt = 'POLYGON (({}))'.format(','.join(coords_poly))
    return wkt


def polygon_to_lonlat(polygon):
    """Convert a WKT polygon string to a tuple of lat and lon lists.

    Parameters
    ----------
    polygon : str
        A WKT polygon string.

    Returns
    -------
    tuple of lists
        A 2-tuple containing a list of longitudes and a list of latitudes.
    """
    poly_coords = polygon.split('((')[1].split('))')[0].split(',')
    coords = [(float(lon), float(lat)) for lon, lat in
              [co.split(' ') for co in poly_coords]]
    lon, lat = zip(*coords)
    return (lon, lat)


def polygon_area(polygon):
    """Computes the area of a WKT polygon in square kilometers.

    The area computation is done by projecting onto an appropriate equal-area
    projection. It is not exact.

    Parameters
    ----------
    polygon : str
        A WKT polygon string.

    Returns
    -------
    float
        The area of the polygon in square kilometers.
    """
    if not PYPROJ_INSTALLED:
        raise ImportError("`pyproj` must be installed to use this feature!")
    poly = wkt_loads(polygon)
    poly_area = shapely.ops.transform(
        partial(
            pyproj.transform,
            pyproj.Proj(init='EPSG:4326'),
            pyproj.Proj(
                proj='aea',
                lat1=poly.bounds[1],
                lat2=poly.bounds[3]
            )
        ),
        poly
    )
    return poly_area.area / 1e6


def intersect(wkt1, wkt2, tolerance=0):
    """Check whether two geometries specified in WKT format intersect.

    Parameters
    ----------
    wkt1 : str
    wkt2 : str

    Returns
    -------
    bool
        True if the geometries intersect, False otherwise.
    """
    poly1 = wkt_loads(wkt1)
    poly2 = wkt_loads(wkt2)
    if tolerance > 0:
        return poly1.distance(poly2) <= tolerance
    else:
        return poly1.intersects(poly2)



#!/usr/bin/env python
# coding=utf-8
""" This is the command line tool for the esahub python package.
"""
from __future__ import print_function
import locale
import warnings
import sys
import argparse
import time
import logging
import datetime
from .config import CONFIG


locale.setlocale(locale.LC_ALL, '')
warnings.filterwarnings("ignore",
                        message="WARNING: unsupported Compound type, skipping")
logger = logging.getLogger('esahub')


# -----------------------------------------------------------------------------
# PARSING COMMAND LINE OPTIONS
# -----------------------------------------------------------------------------
def parse_cli_options(args=None):
    """ Dedicated function for parsing command line options for esahub.
        An array of CLI arguments can be passed to simulate this behavior.
    """

    parser = argparse.ArgumentParser(
        description='Download satellite data archives.',
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
                prog, max_help_position=40))

    subparsers = parser.add_subparsers(help='sub-command help', dest='cmd')
    subparsers.required = True

    # Prepare different program streams:
    parser_get = subparsers.add_parser(
        'get', description='Download satellite data archives.',
        help='Download Sentinel data. \n  Type esahub get --help for details.',
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
                prog, max_help_position=40))

    parser_ls = subparsers.add_parser(
        'ls', description='Query satellite data archives for size.',
        help="Investigate the amount of available data.\n"
             "  Type esahub ls --help for details.",
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
                prog, max_help_position=40))

    parser_doctor = subparsers.add_parser(
        'doctor', description='Check local data for corruption.',
        help="Check already downloaded archives for consistency.\n"
             "  Type esahub doctor --help for details.",
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
                prog, max_help_position=40))

    # ARGUMENTS FOR ALL COMMANDS
    # -------------------------------------------------------------------------
    for p in (parser_get, parser_ls, parser_doctor):
        p.add_argument('--quiet', action='store_true',
                       help='Suppress terminal output.')
        p.add_argument('--log', action='store_true',
                       help='Write a log file.')
        p.add_argument('--debug', action='store_true',
                       help='Enable debugging.')
        p.add_argument('--email', action='store_true',
                       help='Send an email report.')

    # ARGUMENTS FOR GET AND DOCTOR
    # -------------------------------------------------------------------------
    for p in (parser_get, parser_doctor):
        p.add_argument('-d', '--dir',
                       help='Specify the local data directory.')

    # ARGUMENTS FOR GET AND LS
    # -------------------------------------------------------------------------
    for p in (parser_get, parser_ls):
        p.add_argument(
            'sat', nargs='?', default=None,
            choices=list(CONFIG['SATELLITES'].keys()),
            help='Satellite')
        p.add_argument(
            '-m', '--mission',
            help="Select the mission (Sentinel-1, Sentinel-2, Sentinel-3).")
        p.add_argument(
            '--type',
            help="Select products matching specified product type (e.g. GRD).")
        p.add_argument(
            '--orbit',
            help="Orbit direction (ASC, DESC).")
        p.add_argument(
            '--id',
            help="Product identifier, may include wildcards (*).")
        p.add_argument(
            '--server', choices=list(CONFIG['SERVERS'].keys()),
            help="Specify the dowload server.")
        p.add_argument(
            '-g', '--geo', action='append',
            help="Specify geospatial location for intersection as point "
                 "or polygon. You can add multiple options.\n"
                 "  Examples: 'POINT(-6.26 53.35)' (Dublin) \n"
                 "  'POLYGON((-25.1 46.8,-5.25 46.8,-5.25 57.4,"
                 "-25.1 57.4,-25.1 46.8))' (Real Ireland)")
        p.add_argument(
            '--location', action='append',
            choices=[l for l, value in CONFIG['LOCATIONS'].items()],
            help="Specify location semantically as defined in config "
                 "file. You can add multiple options.\n"
                 "  Examples: 'Ireland_Mace_Head'")
        p.add_argument(
            '-t', '--time',
            help="Shortcuts for specifying time intervals. "
                 "Pass any date or date range in string format.\n"
                 "Special options include:\n"
                 "  today|yesterday|24h|midnight")
        p.add_argument(
            '-q', '--query', help='Specify custom query.')

    # ARGUMENTS FOR GET ONLY
    # -------------------------------------------------------------------------
    for p in (parser_get,):
        p.add_argument(
            '-i', '--in', help='Read download list from file.')
        p.add_argument(
            '--restart', action='store_true',
            help='Force restart of incomplete downloads (do not continue).')

    # ARGUMENTS FOR LS ONLY
    # -------------------------------------------------------------------------
    for p in (parser_ls,):
        p.add_argument(
            '-o', '--out', help='Write list to file.')

    # ARGUMENTS FOR DOCTOR ONLY
    # -------------------------------------------------------------------------
    for p in (parser_doctor,):
        p.add_argument(
            '--mode',
            help="Specify the mode for file checking. Options include:\n"
                 "  file - check if archives are valid zip or "
                 "netcdf files (very fast).\n"
                 "  md5 - check if archives match MD5 sum provided online "
                 "(can be slow).")
        p.add_argument(
            '--repair', action='store_true',
            help='Redownload corrupt files.')
        p.add_argument(
            '--delete', action='store_true',
            help='Delete corrupt files. Ignored if running with --repair')

    if args is None:
        return vars(parser.parse_args())
    else:
        return vars(parser.parse_args(args))


# -----------------------------------------------------------------------------
def not_none(dictionary, key):
    return (key in dictionary and dictionary[key] is not None)


def set_config(args):
    global CONFIG
    if not_none(args, 'mode'):
        CONFIG['GENERAL']['CHECK_MODE'] = args['mode']
    if not_none(args, 'force'):
        CONFIG['GENERAL']['SKIP_EXISTING'] = not args['force']
    if not_none(args, 'quiet'):
        CONFIG['GENERAL']['QUIET'] = args['quiet']
    if not_none(args, 'dir'):
        CONFIG['GENERAL']['DAT_DIR'] = args['dir']
    if not_none(args, 'email'):
        CONFIG['GENERAL']['SEND_EMAIL'] = args['email']
    if not_none(args, 'log'):
        CONFIG['GENERAL']['LOGGING'] = args['log']

    if not_none(args, 'sat'):
        CONFIG['GENERAL']['QUERY']['satellite'] = args['sat']
    if not_none(args, 'server'):
        CONFIG['GENERAL']['QUERY']['server'] = args['server']
    if not_none(args, 'mission'):
        CONFIG['GENERAL']['QUERY']['mission'] = args['mission']
    if not_none(args, 'geo'):
        CONFIG['GENERAL']['QUERY']['geo'] = args['geo']
    if not_none(args, 'location'):
        CONFIG['GENERAL']['QUERY']['location'] = args['location']
    if not_none(args, 'time'):
        CONFIG['GENERAL']['QUERY']['time'] = args['time']
    if not_none(args, 'type'):
        CONFIG['GENERAL']['QUERY']['type'] = args['type']
    if not_none(args, 'orbit'):
        CONFIG['GENERAL']['QUERY']['orbit'] = args['orbit']
    if not_none(args, 'id'):
        CONFIG['GENERAL']['QUERY']['id'] = args['id']
    if not_none(args, 'query'):
        CONFIG['GENERAL']['QUERY']['query'] = args['query']

    if not_none(args, 'in'):
        CONFIG['GENERAL']['IN_FILE'] = args['in']
    if not_none(args, 'out'):
        CONFIG['GENERAL']['OUT_FILE'] = args['out']
    if not_none(args, 'restart') and args['restart']:
        CONFIG['GENERAL']['CONTINUE'] = False


# -----------------------------------------------------------------------------
def cli_main():
    """ Execute selected command. """
    args = parse_cli_options()
    set_config(args)
    cmd = args['cmd']

    # This needs to happen before logging is enabled!
    CONFIG['GENERAL']['LOG_FILE'] = CONFIG['GENERAL']['LOG_FILE'].format(
        cmd=cmd,
        time=datetime.datetime.strftime(datetime.datetime.now(),
                                        '%Y-%m-%d_%H-%M-%S')
    )
    CONFIG['GENERAL']['EMAIL_SUBJECT'] = \
        CONFIG['GENERAL']['EMAIL_SUBJECT'].format(cmd=cmd)

    if CONFIG['GENERAL']['SEND_EMAIL']:
        CONFIG['GENERAL']['LOGGING'] = True

    if CONFIG['GENERAL']['LOGGING'] or args['debug']:
        logging.basicConfig(
            filename=CONFIG['GENERAL']['LOG_FILE'],
            format='[%(asctime)s] %(levelname)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.DEBUG if args['debug'] else logging.INFO
        )
    else:
        logging.disable(logging.CRITICAL)

    program = 'esahub {}'.format(' '.join(sys.argv[1:]))
    logger.info('Executing: {}'.format(program))

    #
    # These modules MUST be imported AFTER altering the CONFIG.
    #
    from . import tty, main

    def shutdown():
        del tty.screen

    def interrupt():
        msg = 'Execution interrupted manually.'
        logger.warning(msg)
        tty.screen.result(tty.error(msg))
        shutdown()
        sys.exit()

    try:
        a = time.time()

        if cmd == 'doctor':
            main.doctor(delete=args['delete'], repair=args['repair'])

        elif cmd == 'ls':
            main.ls()

        elif cmd == 'get':
            main.get()

    except KeyboardInterrupt:
        interrupt()

    finally:
        b = time.time()
        logger.info('TOTAL EXECUTION TIME: {0:.1f}s'.format(b-a))

    # Send email report.
    if CONFIG['GENERAL']['LOGGING'] and CONFIG['GENERAL']['SEND_EMAIL'] and \
            len(CONFIG['GENERAL']['EMAIL_REPORT_RECIPIENTS']):
        main.email()

    shutdown()


if __name__ == '__main__':
    cli_main()



# coding=utf-8
""" This module contains methods to check whether a local archive is corrupted.
    It can retrieve the true MD5 from SciHub and compare with the local file
    or check whether the archive is a valid zip file.
"""
from __future__ import print_function
from . import scihub, checksum, tty
from .config import CONFIG
import zipfile
import os
import sys
import logging
try:
    from netCDF4 import Dataset
    NETCDF_INSTALLED = True
except ImportError:
    NETCDF_INSTALLED = False

logger = logging.getLogger('esahub')
PY2 = sys.version_info < (3, 0)
try:
    ZIP_ERROR = zipfile.BadZipFile
except AttributeError:
    ZIP_ERROR = zipfile.BadZipfile


# -----------------------------------------------------------------------------
# FILE LISTING
# -----------------------------------------------------------------------------
def local_ls():
    #
    # Collect a list of all files.
    #
    all_files = []
    for (root, subdirs, files) in os.walk(CONFIG['GENERAL']['TMP_DIR']):
        for f in files:
            full_file_path = os.path.join(root, f)
            if f.endswith('.zip') or f.endswith('.nc'):
                all_files.append(full_file_path)
    return all_files


def check_file(full_file_path, mode):
    return scihub.block(_check_file, full_file_path, mode)


async def _check_file(full_file_path, mode):
    """ Check an already downloaded file for consistency.

    Parameters
    ----------
    full_file_path : str
    mode : {'file', 'md5'}
        If `md5`, check if the md5 checksum matches the value stored on SciHub.
        This is safer, but potentially slow. Also, the md5 checksum may not be
        available for older files. If `file`, check if the archive is a valid
        zip archive or a valid netCDF file. This method is very fast and should
        be accurate in the case of interrupted downloads.

    Returns
    -------
    tuple (str, bool, str)
        A tuple containing the file path, a boolean whether or not the file is
        okay, and a status message.
    """
    message = None

    pbar_key = os.path.split(full_file_path)[1]

    if mode == 'file':
        #
        # Check if the archive is a valid zip archive or
        # a valid netCDF file
        #
        ext = os.path.splitext(full_file_path)[1]
        if ext == '.zip':
            try:
                zip_ref = zipfile.ZipFile(full_file_path, 'r')
                zip_ref.close()
            except ZIP_ERROR as e:
                message = tty.error('BAD ZIP FILE')
                healthy = False
            else:
                message = tty.success('OKAY')
                healthy = True
        elif ext == '.nc':
            if not NETCDF_INSTALLED:
                raise ImportError("`netCDF4` must be installed to use this "
                                  "feature!")
            try:
                nc_ref = Dataset(full_file_path, 'r')
                nc_ref.close()
            except (OSError, IOError) as e:
                message = tty.error('BAD NETCDF FILE')
                healthy = False
            else:
                message = tty.success('OKAY')
                healthy = True
        else:
            message = tty.error('UNKNOWN FILE FORMAT')
            healthy = False

    elif mode == 'md5':
        #
        # Check the md5sum against SciHub.
        #
        remote_md5 = await scihub._md5(product=full_file_path)
        if remote_md5 is False:
            message = tty.error('MD5 NOT FOUND')
            healthy = False
        else:
            if checksum.md5(full_file_path) == remote_md5:
                message = tty.success('MD5 OKAY')
                healthy = True
            else:
                message = tty.error('FILE CORRUPT')
                healthy = False

    result = (full_file_path, healthy, message)

    tty.screen[pbar_key] = (message + ': {name}', tty.NOBAR)
    tty.screen.status(progress=1)

    return result



# import distutils.cmd
import setuptools
import setuptools.command.build_py
from distutils.command.clean import clean as Clean
import os
import shutil


def init_config():
    src = 'esahub/config.yaml'
    dst = os.path.expanduser('~/.esahub.conf')
    if not os.path.isfile(dst):
        shutil.copyfile(src, dst)


class BuildPyCommand(setuptools.command.build_py.build_py):
    """Custom build command."""

    def run(self):
        init_config()
        setuptools.command.build_py.build_py.run(self)


class CleanCommand(Clean):
    description = "Remove build artifacts from the source tree"

    def run(self):
        Clean.run(self)
        # Remove c files if we are not within a sdist package
        cwd = os.path.abspath(os.path.dirname(__file__))
        remove_c_files = not os.path.exists(os.path.join(cwd, 'PKG-INFO'))
        if os.path.exists('build'):
            shutil.rmtree('build')
        for root, dirs, files in os.walk('esahub'):
            for filename in files:
                if any(filename.endswith(suffix) for suffix in
                       (".so", ".pyd", ".dll", ".pyc")):
                    os.unlink(os.path.join(root, filename))
                    continue
                extension = os.path.splitext(filename)[1]
                if remove_c_files and extension in ['.c', '.cpp']:
                    pyx_file = str.replace(filename, extension, '.pyx')
                    if os.path.exists(os.path.join(root, pyx_file)):
                        os.unlink(os.path.join(root, filename))
            for dirname in dirs:
                if dirname == '__pycache__':
                    shutil.rmtree(os.path.join(root, dirname))


cmdclass = {
    'clean': CleanCommand,
    'build_py': BuildPyCommand,
}

setuptools.setup(
    use_scm_version=True,
    cmdclass=cmdclass
)



# coding=utf-8
""" This module implements the main commands accessible from the command line.
"""
from __future__ import print_function
import subprocess
import logging
import sys
import os
import json
import asyncio
from .config import CONFIG
from . import scihub, check, tty, utils

logger = logging.getLogger('esahub')
PY2 = sys.version_info < (3, 0)


# -----------------------------------------------------------------------------
# HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def list_local_archives():
    #
    # Collect a list of all files.
    #
    return utils.ls(CONFIG['GENERAL']['DATA_DIR'])


# -----------------------------------------------------------------------------
# MAIN COMMANDS
# -----------------------------------------------------------------------------
def query_file_list(query=None, limit=None):
    if 'IN_FILE' in CONFIG['GENERAL'] and \
            CONFIG['GENERAL']['IN_FILE'] is not None:
        with open(CONFIG['GENERAL']['IN_FILE'], 'r') as f:
            file_list = json.load(f)
    else:
        if query is None:
            query = CONFIG['GENERAL']['QUERY']
        file_list = scihub.search(query, limit=limit, verbose=True)

    return file_list


def get(query=None, limit=None):
    """This method downloads the files found in a query.

    The main method in the script. Based on the specified command line options
    it queries the webserver, extracts all matching files and downloads them
    into the requested location.

    Parameters
    ----------
    query : dict, optional
        A Query object to restrict the files retrieved.
    limit : int, optional
        The maximum number of files to download.
    """
    file_list = query_file_list(query, limit=limit)
    size = sum(f['size'] for f in file_list)

    msg = 'Downloading {0:d} files ({1}) into {2} ...'.format(
        len(file_list),
        utils.b2h(size),
        CONFIG['GENERAL']['DATA_DIR']
    )
    logging.info(msg)
    tty.screen.status(desc=msg, total=size, mode='bar', reset=True,
                      unit='B', scale=True)

    scihub.download(file_list)


def ls(query=None, quiet=False):
    """List and count files matching the query and compute total file size.

    Parameters
    ----------
    query : dict, optional
        (default: None)
    quiet : bool, optional
        Whether to suppress console output.
    """
    tty.screen.status('Searching ...', mode='static')
    if query is None:
        query = CONFIG['GENERAL']['QUERY']
    file_list = scihub.search(query, verbose=True)
    size = 0.0
    for f in file_list:
        size += f['size']
    if not quiet:
        msg = 'Found {0:d} files ({1}).'.format(len(file_list),
                                                utils.b2h(size))
        logging.info(msg)
        tty.screen.result(msg)
        for f in file_list:
            msg = '{:>8} {}'.format(utils.b2h(f['size']), f['filename'])
            # tty.update(f['filename'],msg)
            logging.info(f['filename'])

    #
    # Write file_list to JSON file
    # so it can be read later by the get() and store() commands.
    #
    if 'OUT_FILE' in CONFIG['GENERAL'] and \
            CONFIG['GENERAL']['OUT_FILE'] is not None:
        with open(CONFIG['GENERAL']['OUT_FILE'], 'w') as f:
            json.dump(file_list, f, default=str, indent=2)

    return file_list


def doctor(delete=False, repair=False):
    """Checks all files in directory for consistency and generates report.

    Parameters
    ----------
    delete : bool, optional
        Whether to delete corrupt files (default: False)
    reapir : bool, optional
        Whether to attempt a redownload of corrupt files (default: False)
    """
    # check._init_bad_file_counter()
    all_files = list_local_archives()
    msg = 'Checking {:d} files for consistency (mode: {}).'.format(
            len(all_files), CONFIG['GENERAL']['CHECK_MODE'])
    logging.info(msg)

    tty.screen.status(desc=msg, reset=True, mode='bar', total=len(all_files),
                      unit='', scale=False)

    #
    # Check every file in the list.
    #
    tasks = []
    for f in all_files:
        tasks.append(
            check._check_file(f, CONFIG['GENERAL']['CHECK_MODE'])
        )
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(asyncio.gather(*tasks))
    bad_files = [status[0] for status in result if status[1] is False]
    n_bad_files = len(bad_files)
    msg = '{0:d}/{1:d} files corrupt.'.format(
        n_bad_files, len(all_files))
    logging.info(msg)
    tty.screen.result(msg)

    if repair:
        # tty.screen.status(desc='Redownloading {} corrupt files ...'
        #                        .format(check.BAD_FILE_COUNTER),
        #                   total=check.BAD_FILE_COUNTER,
        #                   bar=True, unit='', scale=False,
        #                   reset=True)

        scihub.redownload(bad_files)

    elif delete:
        # tty.screen.status(desc='Deleting {} corrupt files ...'
        #                        .format(check.BAD_FILE_COUNTER),
        #                   total=check.BAD_FILE_COUNTER,
        #                   bar=True, unit='', scale=False,
        #                   reset=True)

        for f in bad_files:
            os.remove(f)
            # tty.screen.status(progress=1)

        msg = 'Deleted {} corrupt files!'.format(n_bad_files)
        logging.info(msg)
        tty.screen.result(msg)

    return result


def _product_file(f):
    product = os.path.split(f)[1]
    if product.endswith('.nc.zip'):
        product = os.path.splitext(product)[0]
    else:
        product = os.path.splitext(product)[0]
        sat = utils.get_satellite(product)
        if sat in CONFIG['SATELLITES']:
            product += CONFIG['SATELLITES'][sat]['ext']
        else:
            product += '.zip'
    return product


def email():
    """Send an email report with the current log file.

    This method sends the current log file to the email recipients as specified
    in the config module. It is not meant to be called independently from the
    command line.
    """
    sender = CONFIG['GENERAL']['EMAIL_SENDER']
    recipients = CONFIG['GENERAL']['EMAIL_REPORT_RECIPIENTS']

    from email.mime.text import MIMEText
    with open(CONFIG['GENERAL']['LOG_FILE'], 'rb') as fp:
        log_content = fp.read()
        if not PY2:
            log_content = log_content.decode()
        message = MIMEText(log_content)

    message['Subject'] = CONFIG['GENERAL']['EMAIL_SUBJECT']
    message['From'] = sender
    message['To'] = ','.join(recipients)

    p = subprocess.Popen(
        ['/usr/sbin/sendmail', '-t'],
        stdin=subprocess.PIPE, stdout=subprocess.PIPE,
    )
    if PY2:
        p.communicate(message.as_string())
    else:
        p.communicate(message.as_bytes())
